@@ -7,7 +7,7 @@
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
- * Free Software Foundation; either version 2 of the License, or (at your
+ * Free Software Foundation; either version 3 of the License, or (at your
  * option) any later version.
  *
  * This program is distributed in the hope that it will be useful, but
@@ -63,6 +63,15 @@ static void gucharmap_chartable_finalize   (GObject *object);
 
 static guint signals[NUM_SIGNALS];
 
+#define DEFAULT_FONT_SIZE (20.0 * (double) PANGO_SCALE)
+
+/* These are chosen for compatibility with the older code that
+ * didn't scale the font size by resolution and used 3 and 2.5 here, resp.
+ * Where exactly these factors came from, I don't know.
+ */
+#define FACTOR_WIDTH (2.25) /* 3 / (96 / 72) */
+#define FACTOR_HEIGHT (1.875) /* 2.5 / (96 / 72) */
+
 /** Notes
  *
  * 1. Table geometry
@@ -121,7 +130,6 @@ gucharmap_chartable_set_font_desc_internal (GucharmapChartable *chartable,
                                             PangoFontDescription *font_desc /* adopting */)
 {
   GucharmapChartablePrivate *priv = chartable->priv;
-  int font_size;
 
   if (priv->font_desc)
     pango_font_description_free (priv->font_desc);
@@ -131,14 +139,6 @@ gucharmap_chartable_set_font_desc_internal (GucharmapChartable *chartable,
   if (priv->pango_layout)
     pango_layout_set_font_description (priv->pango_layout, font_desc);
 
-  /* FIXMEchpe: check pango_font_description_get_size_is_absolute() ! */
-  font_size = pango_font_description_get_size (priv->font_desc);
-  /* FIXMEchpe bug 329481 */
-  priv->bare_minimal_column_width = PANGO_PIXELS (3.0 * font_size);
-  priv->bare_minimal_row_height = PANGO_PIXELS (2.5 * font_size);
-
-  priv->drag_font_size = 5 * ((font_size > 0) ? font_size : 10 * PANGO_SCALE);
-
   gtk_widget_queue_resize (GTK_WIDGET (chartable));
 
   g_object_notify (G_OBJECT (chartable), "font-desc");
@@ -263,7 +263,7 @@ position_rectangle_on_screen (GtkWidget *widget,
   
   direction = gtk_widget_get_direction (widget);
   screen = gtk_widget_get_screen (widget);
-  monitor_num = gdk_screen_get_monitor_at_window (screen, widget->window);
+  monitor_num = gdk_screen_get_monitor_at_window (screen, gtk_widget_get_window (widget));
   if (monitor_num < 0)
     monitor_num = 0;
   gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);
@@ -286,7 +286,7 @@ get_root_coords_at_active_char (GucharmapChartable *chartable,
   gint x, y;
   gint row, col;
 
-  gdk_window_get_origin (widget->window, &x, &y);
+  gdk_window_get_origin (gtk_widget_get_window (widget), &x, &y);
 
   row = (priv->active_cell - priv->page_first_cell) / priv->cols;
   col = _gucharmap_chartable_cell_column (chartable, priv->active_cell);
@@ -452,27 +452,6 @@ set_top_row (GucharmapChartable *chartable,
   g_object_notify (G_OBJECT (chartable), "active-character");
 }
 
-static gint
-compute_zoom_font_size (GucharmapChartable *chartable)
-{
-  GucharmapChartablePrivate *priv = chartable->priv;
-  GtkWidget *widget = GTK_WIDGET (chartable);
-  gint screen_height;
-  gdouble limit;
-  gdouble scale;
-  gint font_size;
-
-  screen_height = gdk_screen_get_height (gtk_widget_get_screen (widget));
-
-  limit = (0.3 * screen_height) / priv->bare_minimal_row_height;
-  scale = CLAMP (limit, 1.0, 12.0);
-
-  font_size = pango_font_description_get_size (priv->font_desc);
-  /* FIXMEchpe absolute size ? */
-
-  return scale * ((font_size > 0) ? font_size : 10 * PANGO_SCALE);
-}
-
 /* returns the font family of the last glyph item in the first line of the
  * layout; should be freed by caller */
 static gchar *
@@ -513,8 +492,8 @@ get_font (PangoLayout *layout)
 static PangoLayout *
 layout_scaled_glyph (GucharmapChartable *chartable, 
                      gunichar uc, 
-                     gint font_size,
-                     gchar **font_family)
+                     double font_factor,
+                     char **font_family)
 {
   GucharmapChartablePrivate *priv = chartable->priv;
   PangoFontDescription *font_desc;
@@ -522,7 +501,13 @@ layout_scaled_glyph (GucharmapChartable *chartable,
   gchar buf[11];
 
   font_desc = pango_font_description_copy (priv->font_desc);
-  pango_font_description_set_size (font_desc, font_size);
+
+  if (pango_font_description_get_size_is_absolute (priv->font_desc))
+    pango_font_description_set_absolute_size (font_desc,
+                                              font_factor * pango_font_description_get_size (priv->font_desc));
+  else
+    pango_font_description_set_size (font_desc,
+                                     font_factor * pango_font_description_get_size (priv->font_desc));
 
   layout = pango_layout_new (pango_layout_get_context (priv->pango_layout));
 
@@ -542,7 +527,7 @@ layout_scaled_glyph (GucharmapChartable *chartable,
 static GdkPixmap *
 create_glyph_pixmap (GucharmapChartable *chartable,
                      gunichar wc,
-                     gint font_size,
+                     double font_factor,
                      gboolean draw_font_family)
 {
   GtkWidget *widget = GTK_WIDGET (chartable);
@@ -553,13 +538,13 @@ create_glyph_pixmap (GucharmapChartable *chartable,
   gint pixmap_width, pixmap_height;
   GtkStyle *style;
   GdkPixmap *pixmap;
-  gchar *family;
+  char *family;
 
   /* Apply the scaling.  Unfortunately not all fonts seem to be scalable.
    * We could fall back to GdkPixbuf scaling, but that looks butt ugly :-/
    */
   pango_layout = layout_scaled_glyph (chartable, wc,
-                                      font_size, &family);
+                                      font_factor, &family);
   pango_layout_get_pixel_extents (pango_layout, &char_rect, NULL);
 
   if (draw_font_family)
@@ -583,7 +568,7 @@ create_glyph_pixmap (GucharmapChartable *chartable,
 
   style = gtk_widget_get_style (widget);
 
-  pixmap = gdk_pixmap_new (widget->window,
+  pixmap = gdk_pixmap_new (gtk_widget_get_window (widget),
                            pixmap_width, pixmap_height, -1);
 
   gdk_draw_rectangle (pixmap, style->base_gc[GTK_STATE_NORMAL],
@@ -663,15 +648,51 @@ place_zoom_window_on_active_cell (GucharmapChartable *chartable)
   gtk_window_move (GTK_WINDOW (priv->zoom_window), rect.x, rect.y);
 }
 
+static int
+get_font_size_px (GucharmapChartable *chartable)
+{
+  GucharmapChartablePrivate *priv = chartable->priv;
+  GtkWidget *widget = GTK_WIDGET (chartable);
+  GdkScreen *screen;
+  double resolution;
+  int font_size;
+
+  g_assert (priv->font_desc != NULL);
+
+  screen = gtk_widget_get_screen (widget);
+  resolution = gdk_screen_get_resolution (screen);
+  if (resolution < 0.0) /* will be -1 if the resolution is not defined in the GdkScreen */
+    resolution = 96.0;
+
+  if (pango_font_description_get_size_is_absolute (priv->font_desc))
+    font_size = pango_font_description_get_size (priv->font_desc);
+  else
+    font_size = ((double) pango_font_description_get_size (priv->font_desc)) * resolution / 72.0;
+
+  if (PANGO_PIXELS (font_size) <= 0)
+    font_size = DEFAULT_FONT_SIZE * resolution / 72.0;
+
+  return PANGO_PIXELS (font_size);
+}
+
 static void
 update_zoom_window (GucharmapChartable *chartable)
 {
   GucharmapChartablePrivate *priv = chartable->priv;
+  GtkWidget *widget = GTK_WIDGET (chartable);
   GdkPixmap *pixmap;
+  double scale;
+  int font_size_px, screen_height;
+
+  font_size_px = get_font_size_px (chartable);
+  screen_height = gdk_screen_get_height (gtk_widget_get_screen (widget));
+
+  scale = (0.3 * screen_height) / (FACTOR_WIDTH * font_size_px);
+  scale = CLAMP (scale, 1.0, 12.0);
 
   pixmap = create_glyph_pixmap (chartable,
                                 gucharmap_chartable_get_active_character (chartable),
-                                compute_zoom_font_size (chartable),
+                                scale,
                                 TRUE);
   gtk_image_set_from_pixmap (GTK_IMAGE (priv->zoom_image), pixmap, NULL);
   g_object_unref (pixmap);
@@ -712,7 +733,7 @@ destroy_zoom_window (GucharmapChartable *chartable)
       priv->zoom_window = NULL;
       priv->zoom_image = NULL;
 
-      gdk_window_set_cursor (widget->window, NULL);
+      gdk_window_set_cursor (gtk_widget_get_window (widget), NULL);
       gtk_widget_destroy (zoom_window);
     }
 }
@@ -812,33 +833,45 @@ draw_borders (GucharmapChartable *chartable)
 {
   GucharmapChartablePrivate *priv = chartable->priv;
   GtkWidget *widget = GTK_WIDGET (chartable);
+  GtkAllocation *allocation;
+  GtkStyle *style;
   gint x, y, col, row;
+#if GTK_CHECK_VERSION (2, 18, 0)
+  GtkAllocation widget_allocation;
+
+  gtk_widget_get_allocation (widget, &widget_allocation);
+  allocation = &widget_allocation;
+#else
+  allocation = &widget->allocation;
+#endif
 
   /* dark_gc[GTK_STATE_NORMAL] seems to be what is used to draw the borders
    * around widgets, so we use it for the lines */
 
+  style = gtk_widget_get_style (widget);
+
   /* vertical lines */
   gdk_draw_line (priv->pixmap,
-                 widget->style->dark_gc[GTK_STATE_NORMAL],
-                 0, 0, 0, widget->allocation.height - 1);
+                 style->dark_gc[GTK_STATE_NORMAL],
+                 0, 0, 0, allocation->height - 1);
   for (col = 0, x = 0;  col < priv->cols;  col++)
     {
       x += _gucharmap_chartable_column_width (chartable, col);
       gdk_draw_line (priv->pixmap,
-                     widget->style->dark_gc[GTK_STATE_NORMAL],
-                     x, 0, x, widget->allocation.height - 1);
+                     style->dark_gc[GTK_STATE_NORMAL],
+                     x, 0, x, allocation->height - 1);
     }
 
   /* horizontal lines */
   gdk_draw_line (priv->pixmap,
-                 widget->style->dark_gc[GTK_STATE_NORMAL],
-                 0, 0, widget->allocation.width - 1, 0);
+                 style->dark_gc[GTK_STATE_NORMAL],
+                 0, 0, allocation->width - 1, 0);
   for (row = 0, y = 0;  row < priv->rows;  row++)
     {
       y += _gucharmap_chartable_row_height (chartable, row);
       gdk_draw_line (priv->pixmap,
-                     widget->style->dark_gc[GTK_STATE_NORMAL],
-                     0, y, widget->allocation.width - 1, y);
+                     style->dark_gc[GTK_STATE_NORMAL],
+                     0, y, allocation->width - 1, y);
     }
 }
 
@@ -896,6 +929,7 @@ draw_character (GucharmapChartable *chartable,
   gunichar wc;
   guint cell;
   GdkGC *gc;
+  GtkStyle *style;
   gchar buf[10];
   gint n;
 
@@ -905,12 +939,18 @@ draw_character (GucharmapChartable *chartable,
   if (wc > UNICHAR_MAX || !gucharmap_unichar_validate (wc) || !gucharmap_unichar_isdefined (wc))
     return;
 
+  style = gtk_widget_get_style (widget);
+
+#if GTK_CHECK_VERSION (2, 18, 0)
+  if (gtk_widget_has_focus (widget) && (gint)cell == priv->active_cell)
+#else
   if (GTK_WIDGET_HAS_FOCUS (widget) && (gint)cell == priv->active_cell)
-    gc = widget->style->text_gc[GTK_STATE_SELECTED];
+#endif
+    gc = style->text_gc[GTK_STATE_SELECTED];
   else if ((gint)cell == priv->active_cell)
-    gc = widget->style->text_gc[GTK_STATE_ACTIVE];
+    gc = style->text_gc[GTK_STATE_ACTIVE];
   else
-    gc = widget->style->text_gc[GTK_STATE_NORMAL];
+    gc = style->text_gc[GTK_STATE_NORMAL];
 
   square_width = _gucharmap_chartable_column_width (chartable, col) - 1;
   square_height = _gucharmap_chartable_row_height (chartable, row) - 1;
@@ -938,26 +978,33 @@ draw_square_bg (GucharmapChartable *chartable, gint row, gint col)
   gint square_width, square_height; 
   GdkGC *gc;
   GdkColor untinted;
+  GtkStyle *style;
   guint cell;
   gunichar wc;
 
   cell = get_cell_at_rowcol (chartable, row, col);
   wc = gucharmap_codepoint_list_get_char (priv->codepoint_list, cell);
 
-  gc = gdk_gc_new (GDK_DRAWABLE (widget->window));
+  gc = gdk_gc_new (GDK_DRAWABLE (gtk_widget_get_window (widget)));
 
+  style = gtk_widget_get_style (widget);
+
+#if GTK_CHECK_VERSION (2, 18, 0)
+  if (gtk_widget_has_focus (widget) && (gint)cell == priv->active_cell)
+#else
   if (GTK_WIDGET_HAS_FOCUS (widget) && (gint)cell == priv->active_cell)
-    untinted = widget->style->base[GTK_STATE_SELECTED];
+#endif
+    untinted = style->base[GTK_STATE_SELECTED];
   else if ((gint)cell == priv->active_cell)
-    untinted = widget->style->base[GTK_STATE_ACTIVE];
+    untinted = style->base[GTK_STATE_ACTIVE];
   else if ((gint)cell > priv->last_cell)
-    untinted = widget->style->dark[GTK_STATE_NORMAL];
+    untinted = style->dark[GTK_STATE_NORMAL];
   else if (! gucharmap_unichar_validate (wc))
-    untinted = widget->style->fg[GTK_STATE_INSENSITIVE];
+    untinted = style->fg[GTK_STATE_INSENSITIVE];
   else if (! gucharmap_unichar_isdefined (wc))
-    untinted = widget->style->bg[GTK_STATE_INSENSITIVE];
+    untinted = style->bg[GTK_STATE_INSENSITIVE];
   else
-    untinted = widget->style->base[GTK_STATE_NORMAL];
+    untinted = style->base[GTK_STATE_NORMAL];
 
   gdk_gc_set_rgb_fg_color (gc, &untinted);
 
@@ -1014,17 +1061,30 @@ draw_chartable_from_scratch (GucharmapChartable *chartable)
   GucharmapChartablePrivate *priv = chartable->priv;
   GtkWidget *widget = GTK_WIDGET (chartable);
   gint row, col;
+  GtkAllocation *allocation;
+#if GTK_CHECK_VERSION (2, 18, 0)
+  GtkAllocation widget_allocation;
+
+  gtk_widget_get_allocation (widget, &widget_allocation);
+  allocation = &widget_allocation;
+#else
+  allocation = &widget->allocation;
+#endif
 
   /* drawing area may not be exposed yet when restoring last char setting
    */
+#if GTK_CHECK_VERSION (2, 20, 0)
+  if (!gtk_widget_get_realized (GTK_WIDGET (chartable)))
+#else
   if (!GTK_WIDGET_REALIZED (chartable))
+#endif
     return;
 
   if (priv->pixmap == NULL)
     priv->pixmap = gdk_pixmap_new (
-	    widget->window, 
-	    widget->allocation.width,
-	    widget->allocation.height, -1);
+	    gtk_widget_get_window (widget),
+	    allocation->width,
+	    allocation->height, -1);
 
   draw_borders (chartable);
 
@@ -1044,6 +1104,18 @@ copy_rows (GucharmapChartable *chartable, gint row_offset)
   GtkWidget *widget = GTK_WIDGET (chartable);
   gint num_padded_rows = priv->n_padded_rows;
   gint from_row, to_row;
+  GtkStyle *style;
+  GtkAllocation *allocation;
+#if GTK_CHECK_VERSION (2, 18, 0)
+  GtkAllocation widget_allocation;
+
+  gtk_widget_get_allocation (widget, &widget_allocation);
+  allocation = &widget_allocation;
+#else
+  allocation = &widget->allocation;
+#endif
+
+  style = gtk_widget_get_style (widget);
 
   if (ABS (row_offset) < priv->rows - num_padded_rows)
     {
@@ -1067,11 +1139,11 @@ copy_rows (GucharmapChartable *chartable, gint row_offset)
 
       gdk_draw_drawable (
               priv->pixmap,
-              widget->style->base_gc[GTK_STATE_NORMAL],
+              style->base_gc[GTK_STATE_NORMAL],
               priv->pixmap,
               0, _gucharmap_chartable_y_offset (chartable, from_row), 
               0, _gucharmap_chartable_y_offset (chartable, to_row),
-              widget->allocation.width, height);
+              allocation->width, height);
     }
 
   if (ABS (row_offset) < num_padded_rows)
@@ -1091,12 +1163,12 @@ copy_rows (GucharmapChartable *chartable, gint row_offset)
       /* it's ok to go off the end (so use allocation.height) */
       gdk_draw_drawable (
               priv->pixmap,
-              widget->style->base_gc[GTK_STATE_NORMAL],
+              style->base_gc[GTK_STATE_NORMAL],
               priv->pixmap,
               0, _gucharmap_chartable_y_offset (chartable, from_row), 
               0, _gucharmap_chartable_y_offset (chartable, to_row),
-              widget->allocation.width, 
-              widget->allocation.height);
+              allocation->width,
+              allocation->height);
     }
 }
 
@@ -1215,14 +1287,15 @@ update_scrollbar_adjustment (GucharmapChartable *chartable)
   if (!vadjustment)
     return;
 
-  vadjustment->value = 1.0 * priv->page_first_cell / priv->cols;
-  vadjustment->lower = 0.0;
-  vadjustment->upper = 1.0 * ( priv->last_cell / priv->cols + 1 );
-  vadjustment->step_increment = 3.0;
-  vadjustment->page_increment = 1.0 * priv->rows;
-  vadjustment->page_size = priv->rows; /* FIXMEchpe + 1 maybe? so scroll-wheel up/down scroll exactly half a page? */
-
-  gtk_adjustment_changed (vadjustment);
+  gtk_adjustment_configure (vadjustment,
+                            1.0 * priv->page_first_cell / priv->cols,
+                            0.0,
+                            1.0 * ( priv->last_cell / priv->cols + 1 ),
+                            3.0,
+                            1.0 * priv->rows,
+                            /* FIXMEchpe: shouldn't set page size at all! */
+                            /* FIXMEchpe + 1 maybe? so scroll-wheel up/down scroll exactly half a page? */
+                            priv->rows);
 }
 
 static void
@@ -1311,12 +1384,19 @@ gucharmap_chartable_drag_begin (GtkWidget *widget,
                                 GdkDragContext *context)
 {
   GucharmapChartable *chartable = GUCHARMAP_CHARTABLE (widget);
-  GucharmapChartablePrivate *priv = chartable->priv;
   GdkPixmap *drag_icon;
+  double scale;
+  int font_size_px, screen_height;
+
+  font_size_px = get_font_size_px (chartable);
+  screen_height = gdk_screen_get_height (gtk_widget_get_screen (widget));
+
+  scale = (0.3 * screen_height) / (FACTOR_WIDTH * font_size_px);
+  scale = CLAMP (scale, 1.0, 5.0);
 
   drag_icon = create_glyph_pixmap (chartable,
                                    gucharmap_chartable_get_active_character (chartable),
-                                   priv->drag_font_size,
+                                   scale,
                                    FALSE);
   gtk_drag_set_icon_pixmap (context, gtk_widget_get_colormap (widget), 
                             drag_icon, NULL, -8, -8);
@@ -1357,7 +1437,8 @@ gucharmap_chartable_drag_data_received (GtkWidget *widget,
   gchar *text;
   gunichar wc;
 
-  if (selection_data->length <= 0 || selection_data->data == NULL)
+  if (gtk_selection_data_get_length (selection_data) <= 0 ||
+      gtk_selection_data_get_data (selection_data) == NULL)
     return;
 
   text = (gchar *) gtk_selection_data_get_text (selection_data);
@@ -1389,9 +1470,17 @@ gucharmap_chartable_expose_event (GtkWidget *widget,
 {
   GucharmapChartable *chartable = GUCHARMAP_CHARTABLE (widget);
   GucharmapChartablePrivate *priv = chartable->priv;
+#if GTK_CHECK_VERSION(2,90,5)
+  cairo_rectangle_int_t *rect;
+  cairo_rectangle_int_t r;
+#else
   GdkRectangle *rects;
+  GdkRectangle *rect;
+#endif
   int i, n_rects;
   GdkGC *gc;
+  GtkStyle *style;
+  GdkWindow *window;
 
   /* Don't draw anything if we haven't set a codepoint list yet */
   if (priv->codepoint_list == NULL)
@@ -1402,10 +1491,16 @@ gucharmap_chartable_expose_event (GtkWidget *widget,
       draw_chartable_from_scratch (chartable);
     }
 
+#if GTK_CHECK_VERSION(2,90,5)
+  if (cairo_region_is_empty (event->region))
+    return FALSE;
+  n_rects = cairo_region_num_rectangles (event->region);
+#else
   if (gdk_region_empty (event->region))
     return FALSE;
-
   gdk_region_get_rectangles (event->region, &rects, &n_rects);
+#endif
+
   if (n_rects == 0)
     return FALSE;
 
@@ -1420,17 +1515,29 @@ gucharmap_chartable_expose_event (GtkWidget *widget,
   }
 #endif
 
-  gc = widget->style->fg_gc[GTK_STATE_NORMAL];
+  style = gtk_widget_get_style (widget);
+  window = gtk_widget_get_window (widget);
+
+  gc = style->fg_gc[GTK_STATE_NORMAL];
   for (i = 0; i < n_rects; ++i) {
-    gdk_draw_drawable (widget->window,
+#if GTK_CHECK_VERSION(2,90,5)
+    cairo_region_get_rectangle (event->region, i, &r);
+    rect = &r;
+#else
+    rect = rects + i;
+#endif
+    gdk_draw_drawable (window,
                        gc,
                        priv->pixmap,
-                       rects[i].x, rects[i].y,
-                       rects[i].x, rects[i].y,
-                       rects[i].width, rects[i].height);
+                       rect->x, rect->y,
+                       rect->x, rect->y,
+                       rect->width, rect->height);
   }
 
+#if GTK_CHECK_VERSION(2,90,5)
+#else
   g_free (rects);
+#endif
 
   /* no need to chain up */
   return FALSE;
@@ -1561,7 +1668,7 @@ gucharmap_chartable_realize (GtkWidget *widget)
 
   GTK_WIDGET_CLASS (gucharmap_chartable_parent_class)->realize (widget);
 
-  gdk_window_set_back_pixmap (widget->window, NULL, FALSE);
+  gdk_window_set_back_pixmap (gtk_widget_get_window (widget), NULL, FALSE);
     
   if (priv->pixmap != NULL)
     {
@@ -1581,18 +1688,36 @@ gucharmap_chartable_size_allocate (GtkWidget *widget,
   int old_rows, old_cols;
   int total_extra_pixels;
   int new_first_cell;
+  int bare_minimal_column_width, bare_minimal_row_height;
+  int font_size_px;
+#if GTK_CHECK_VERSION (2, 18, 0)
+  GtkAllocation widget_allocation;
+#endif
 
   GTK_WIDGET_CLASS (gucharmap_chartable_parent_class)->size_allocate (widget, allocation);
 
+#if GTK_CHECK_VERSION (2, 18, 0)
+  gtk_widget_get_allocation (widget, &widget_allocation);
+  allocation = &widget_allocation;
+#else
+  allocation = &widget->allocation;
+#endif
+
   old_rows = priv->rows;
   old_cols = priv->cols;
 
-  if (priv->snap_pow2_enabled)
-    priv->cols = (1 << g_bit_nth_msf ((allocation->width - 1) / priv->bare_minimal_column_width, -1));
-  else
-    priv->cols = (allocation->width - 1) / priv->bare_minimal_column_width;
+  font_size_px = get_font_size_px (chartable);
 
-  priv->rows = (allocation->height - 1) / priv->bare_minimal_row_height;
+  /* FIXMEchpe bug 329481 */
+  bare_minimal_column_width = FACTOR_WIDTH * font_size_px;
+  bare_minimal_row_height = FACTOR_HEIGHT * font_size_px;
+
+  if (priv->snap_pow2_enabled)
+    priv->cols = (1 << g_bit_nth_msf ((allocation->width - 1) / bare_minimal_column_width, -1));
+  else
+    priv->cols = (allocation->width - 1) / bare_minimal_column_width;
+
+  priv->rows = (allocation->height - 1) / bare_minimal_row_height;
 
   /* avoid a horrible floating point exception crash */
   if (priv->rows < 1)
@@ -1602,13 +1727,13 @@ gucharmap_chartable_size_allocate (GtkWidget *widget,
 
   priv->page_size = priv->rows * priv->cols;
 
-  total_extra_pixels = widget->allocation.width - (priv->cols * priv->bare_minimal_column_width + 1);
-  priv->minimal_column_width = priv->bare_minimal_column_width + total_extra_pixels / priv->cols;
-  priv->n_padded_columns = widget->allocation.width - (priv->minimal_column_width * priv->cols + 1);
+  total_extra_pixels = allocation->width - (priv->cols * bare_minimal_column_width + 1);
+  priv->minimal_column_width = bare_minimal_column_width + total_extra_pixels / priv->cols;
+  priv->n_padded_columns = allocation->width - (priv->minimal_column_width * priv->cols + 1);
 
-  total_extra_pixels = widget->allocation.height - (priv->rows * priv->bare_minimal_row_height + 1);
-  priv->minimal_row_height = priv->bare_minimal_row_height + total_extra_pixels / priv->rows;
-  priv->n_padded_rows = widget->allocation.height - (priv->minimal_row_height * priv->rows + 1);
+  total_extra_pixels = allocation->height - (priv->rows * bare_minimal_row_height + 1);
+  priv->minimal_row_height = bare_minimal_row_height + total_extra_pixels / priv->rows;
+  priv->n_padded_rows = allocation->height - (priv->minimal_row_height * priv->rows + 1);
 
   /* force pixmap to be redrawn on next expose event */
   if (priv->pixmap != NULL)
@@ -1638,10 +1763,12 @@ gucharmap_chartable_size_request (GtkWidget *widget,
                                   GtkRequisition *requisition)
 {
   GucharmapChartable *chartable = GUCHARMAP_CHARTABLE (widget);
-  GucharmapChartablePrivate *priv = chartable->priv;
-  
-  requisition->width = priv->bare_minimal_column_width;
-  requisition->height = priv->bare_minimal_row_height;
+  int font_size_px;
+
+  font_size_px = get_font_size_px (chartable);
+
+  requisition->width = FACTOR_WIDTH * font_size_px;
+  requisition->height = FACTOR_HEIGHT * font_size_px;
 }
 
 static void
@@ -1662,12 +1789,22 @@ gucharmap_chartable_style_set (GtkWidget *widget,
   priv->pango_layout = NULL;
 
   if (priv->font_desc == NULL) {
+    GtkStyle *style;
     PangoFontDescription *font_desc;
 
-    font_desc = pango_font_description_copy (widget->style->font_desc);
-    pango_font_description_set_size (font_desc,
-                                     2.0 * pango_font_description_get_size (font_desc));
-    gucharmap_chartable_set_font_desc_internal (chartable, font_desc);
+    style = gtk_widget_get_style (widget);
+    font_desc = pango_font_description_copy (style->font_desc);
+
+    /* Use twice the size of the style's font */
+    if (pango_font_description_get_size_is_absolute (font_desc))
+      pango_font_description_set_absolute_size (font_desc,
+                                                2 * pango_font_description_get_size (font_desc));
+    else
+      pango_font_description_set_size (font_desc,
+                                       2 * pango_font_description_get_size (font_desc));
+
+    gucharmap_chartable_set_font_desc_internal (chartable, font_desc /* adopts */);
+    g_assert (priv->font_desc != NULL);
   }
 
   priv->pango_layout = gtk_widget_create_pango_layout (widget, NULL);
@@ -1930,7 +2067,11 @@ gucharmap_chartable_paste_clipboard (GucharmapChartable *chartable)
   GtkClipboard *clipboard;
   gpointer *data;
 
+#if GTK_CHECK_VERSION (2, 20, 0)
+  if (!gtk_widget_get_realized (GTK_WIDGET (chartable)))
+#else
   if (!GTK_WIDGET_REALIZED (chartable))
+#endif
     return;
 
   data = g_slice_new (gpointer);
@@ -1979,7 +2120,11 @@ gucharmap_chartable_init (GucharmapChartable *chartable)
   gtk_drag_dest_add_text_targets (widget);
 
   /* this is required to get key_press events */
+#if GTK_CHECK_VERSION (2, 18, 0)
+  gtk_widget_set_can_focus (widget, TRUE);
+#else
   GTK_WIDGET_SET_FLAGS (widget, GTK_CAN_FOCUS);
+#endif
 
   gucharmap_chartable_set_codepoint_list (chartable, NULL);
 
