@@ -3,7 +3,7 @@
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
- * Free Software Foundation; either version 2 of the License, or (at your
+ * Free Software Foundation; either version 3 of the License, or (at your
  * option) any later version.
  *
  * This program is distributed in the hope that it will be useful, but
@@ -555,7 +555,11 @@ information_dialog (GucharmapSearchDialog *search_dialog,
   GucharmapSearchDialogPrivate *priv = GUCHARMAP_SEARCH_DIALOG_GET_PRIVATE (search_dialog);
   GtkWidget *dialog;
 
+#if GTK_CHECK_VERSION (2,18,0)
+  dialog = gtk_message_dialog_new (gtk_widget_get_visible (GTK_WIDGET (search_dialog)) ?
+#else
   dialog = gtk_message_dialog_new (GTK_WIDGET_VISIBLE (search_dialog) ?
+#endif
                                      GTK_WINDOW (search_dialog) :
                                      GTK_WINDOW (priv->guw),
                                    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
@@ -587,7 +591,7 @@ search_completed (GucharmapSearchDialog *search_dialog)
       gtk_widget_set_sensitive (priv->next_button, FALSE);
     }
 
-  gdk_window_set_cursor (GTK_WIDGET (search_dialog)->window, NULL);
+  gdk_window_set_cursor (gtk_widget_get_window (GTK_WIDGET (search_dialog)), NULL);
 }
 
 static gboolean
@@ -616,7 +620,7 @@ _gucharmap_search_dialog_fire_search (GucharmapSearchDialog *search_dialog,
     return;
 
   GdkCursor *cursor = _gucharmap_window_progress_cursor ();
-  gdk_window_set_cursor (GTK_WIDGET (search_dialog)->window, cursor);
+  gdk_window_set_cursor (gtk_widget_get_window (GTK_WIDGET (search_dialog)), cursor);
   gdk_cursor_unref (cursor);
 
   list = gucharmap_charmap_get_book_codepoint_list (priv->guw->charmap);
@@ -723,14 +727,16 @@ static void
 gucharmap_search_dialog_init (GucharmapSearchDialog *search_dialog)
 {
   GucharmapSearchDialogPrivate *priv = GUCHARMAP_SEARCH_DIALOG_GET_PRIVATE (search_dialog);
-  GtkWidget *hbox, *label;
+  GtkWidget *hbox, *label, *content_area;
+
+  content_area = gtk_dialog_get_content_area (GTK_DIALOG (search_dialog));
 
   /* follow hig guidelines */
   gtk_window_set_title (GTK_WINDOW (search_dialog), _("Find"));
   gtk_container_set_border_width (GTK_CONTAINER (search_dialog), 6);
   gtk_dialog_set_has_separator (GTK_DIALOG (search_dialog), FALSE);
   gtk_window_set_destroy_with_parent (GTK_WINDOW (search_dialog), TRUE);
-  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (search_dialog)->vbox), 12);
+  gtk_box_set_spacing (GTK_BOX (content_area), 12);
   gtk_window_set_resizable (GTK_WINDOW (search_dialog), FALSE);
 
   g_signal_connect (search_dialog, "delete-event", G_CALLBACK (gtk_widget_hide_on_delete), NULL);
@@ -739,13 +745,21 @@ gucharmap_search_dialog_init (GucharmapSearchDialog *search_dialog)
   gtk_dialog_add_button (GTK_DIALOG (search_dialog), GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);
 
   priv->prev_button = gtk_button_new ();
+#if GTK_CHECK_VERSION (2,18,0)
+  gtk_widget_set_can_default (priv->prev_button, TRUE);
+#else
   GTK_WIDGET_SET_FLAGS (priv->prev_button, GTK_CAN_DEFAULT);
+#endif
   set_button_stock_image_and_label (GTK_BUTTON (priv->prev_button), GTK_STOCK_GO_BACK, _("_Previous"));
   gtk_dialog_add_action_widget (GTK_DIALOG (search_dialog), priv->prev_button, GUCHARMAP_RESPONSE_PREVIOUS);
   gtk_widget_show (priv->prev_button);
 
   priv->next_button = gtk_button_new ();
-  GTK_WIDGET_SET_FLAGS (priv->next_button, GTK_CAN_DEFAULT);
+#if GTK_CHECK_VERSION (2,18,0)
+  gtk_widget_set_can_default (priv->prev_button, TRUE);
+#else
+  GTK_WIDGET_SET_FLAGS (priv->prev_button, GTK_CAN_DEFAULT);
+#endif
   gtk_widget_show (priv->next_button);
   set_button_stock_image_and_label (GTK_BUTTON (priv->next_button), GTK_STOCK_GO_FORWARD, _("_Next"));
   gtk_dialog_add_action_widget (GTK_DIALOG (search_dialog), priv->next_button, GUCHARMAP_RESPONSE_NEXT);
@@ -760,7 +774,7 @@ gucharmap_search_dialog_init (GucharmapSearchDialog *search_dialog)
   hbox = gtk_hbox_new (FALSE, 12);
   gtk_widget_show (hbox);
   gtk_container_set_border_width (GTK_CONTAINER (hbox), 6);
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (search_dialog)->vbox), hbox, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (content_area), hbox, FALSE, FALSE, 0);
 
   label = gtk_label_new_with_mnemonic (_("_Search:"));
   gtk_widget_show (label);
@@ -774,12 +788,12 @@ gucharmap_search_dialog_init (GucharmapSearchDialog *search_dialog)
 
   priv->whole_word_option = gtk_check_button_new_with_mnemonic (_("Match _whole word"));
   gtk_widget_show (priv->whole_word_option);
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (search_dialog)->vbox), priv->whole_word_option, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (content_area), priv->whole_word_option, FALSE, FALSE, 0);
   g_signal_connect (priv->whole_word_option, "toggled", G_CALLBACK (entry_changed), search_dialog);
 
   priv->annotations_option = gtk_check_button_new_with_mnemonic (_("Search in character _details"));
   gtk_widget_show (priv->annotations_option);
-  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (search_dialog)->vbox), priv->annotations_option, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (content_area), priv->annotations_option, FALSE, FALSE, 0);
   g_signal_connect (priv->annotations_option, "toggled", G_CALLBACK (entry_changed), search_dialog);
 
   gtk_label_set_mnemonic_widget (GTK_LABEL (label), priv->entry);
